from django import template

from ..models import RequestFingerprint

register = template.Library()


@register.simple_tag
def hit_count(url: str) -> int:
    return RequestFingerprint.get_count_for_urls([url])[url]



from contextlib import suppress
from datetime import timedelta
from functools import wraps
from logging import getLogger
from typing import Optional

from django.conf import settings
from django.core.exceptions import BadRequest, DisallowedRedirect
from django.db import transaction
from django.http.response import HttpResponse
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.decorators import method_decorator
from django.utils.encoding import iri_to_uri
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.timezone import now
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import TemplateView
from ipware import get_client_ip

from .models import BrowserFingerprint, RequestFingerprint, Url, UserSession

log = getLogger(__name__)

UTM_PARAMS = (
    "utm_source",
    "utm_campaign",
    "utm_term",
    "utm_medium",
    "utm_content",
)


def get_or_create_session_key(request) -> str:
    if not request.session or not request.session.session_key:
        request.session.create()
    return request.session.session_key


def fingerprint(fn):
    """A decorator which creates a backend Fingerprint object for the current request."""

    @wraps(fn)
    def wrapper(request, *args, **kwargs):
        session_key = get_or_create_session_key(request)
        debounce_period = getattr(settings, "FINGERPRINT_DEBOUNCE_PERIOD", timedelta(seconds=10))

        url_value = request.build_absolute_uri()
        url, _ = Url.objects.get_get_or_create(value=url_value)

        referer = request.META.get("HTTP_REFERER", "")

        session_defaults = {param: request.GET.get(param, "") for param in UTM_PARAMS}
        session_defaults["referer"] = referer

        with suppress(RequestFingerprint.MultipleObjectsReturned), transaction.atomic():
            fingerprint, created = RequestFingerprint.objects.get_or_create(
                user_session=UserSession.objects.get_or_create(session_key=session_key, defaults=session_defaults)[0],
                url=url,
                created__gte=now() - debounce_period,
                defaults=dict(
                    ip=get_client_ip(request)[0],
                    user_agent=request.META.get("HTTP_USER_AGENT", ""),
                    accept=request.META.get("HTTP_ACCEPT", ""),
                    content_encoding=request.META.get("HTTP_CONTENT_ENCODING", ""),
                    content_language=request.META.get("HTTP_CONTENT_LANGUAGE", ""),
                    referer=referer,
                    cf_ipcountry=request.META.get("HTTP_CF_IPCOUNTRY", ""),
                ),
            )
            log.debug("Fingerprint %s, created=%s", fingerprint, created)
        return fn(request, *args, **kwargs)

    return wrapper


def remember_user_session(fn):
    """
    A decorator to match a user to a session.

    When database session backend is used, this is not necessary, as UserSession creation
    is triggered by Session's post_save signal. However, with other session backends this
    signal may not be triggered, thus we decorate some view with this decorator, and it
    remembers UserSession if user is logged in.
    """

    @wraps(fn)
    def wrapper(request, *args, **kwargs):
        if request.user.is_authenticated:
            UserSession.objects.update_or_create(
                session_key=get_or_create_session_key(request),
                defaults=dict(user=request.user),
            )
        return fn(request, *args, **kwargs)

    return wrapper


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(remember_user_session, name="get")
@method_decorator(fingerprint, name="get")
class FingerprintView(TemplateView):
    """
    Multi-purpose fingerprinting view.

    If user makes a GET request to this page, a page with both a server and browser fingerprinting is displayed.
    Server fingerprinting is done when processing GET request in the view (by using @fingerprint decorator),
    and browser fingerprinting is done by javascript code in the template. Javascript code calculates visitor_id
    based on browser fingerprint, and then sends it to this view via POST request.

    If there is `next` parameter in the query string, user is redirected to that page after fingerprinting. Otherwise
    user will be redirected to main page.

    If user sends POST request to this page, it is assumed that the request is coming from javascript code and
    contains `id` post parameter, which is effectively a browser fingerprint (visitor_id).

    If redirect to other domains is required, then this view should be subclassed and
    `allowed_hosts` class variable should contain a set of allowed domains, say, `{'google.com'}`.
    By default only relative paths are allowed.
    """

    template_name = "fingerprint/fingerprint.html"
    redirect_in = timedelta(seconds=3)
    allowed_hosts: Optional[set] = None

    def get_context_data(self, **kwargs):
        redirect_url = iri_to_uri(self.request.GET.get("next", "/"))
        if not url_has_allowed_host_and_scheme(redirect_url, self.allowed_hosts):
            raise DisallowedRedirect()

        return {
            "redirect_in": self.redirect_in,
            "redirect_url": redirect_url,
        }

    def post(self, request, *args, **kwargs):
        try:
            visitor_id = request.POST["id"]
        except MultiValueDictKeyError:
            raise BadRequest()

        session_key = get_or_create_session_key(request)

        url_value = request.META.get("HTTP_REFERER", "")
        url, _ = Url.objects.get_get_or_create(value=url_value)

        with transaction.atomic():
            BrowserFingerprint.objects.create(
                user_session=UserSession.objects.get_or_create(session_key=session_key)[0],
                url=url,
                visitor_id=visitor_id,
            )
        return HttpResponse(status=200)



# Generated by Django 5.0.4 on 2024-04-26 19:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("fingerprint", "0003_alter_browserfingerprint_url_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="requestfingerprint",
            name="cf_ipcountry",
            field=models.CharField(blank=True, max_length=16),
        ),
        migrations.AddField(
            model_name="requestfingerprint",
            name="referer",
            field=models.CharField(blank=True, max_length=2047),
        ),
    ]



from __future__ import annotations

from collections.abc import Iterator
from datetime import timedelta
from functools import wraps
from itertools import chain, islice
from typing import Callable

from django.contrib import admin
from django.db.models import Count, Max, Prefetch, Q
from django.urls import reverse
from django.utils.html import format_html
from django.utils.timezone import now

from .models import BrowserFingerprint, RequestFingerprint, UserFingerprint, UserSession

try:
    from itertools import pairwise
except ImportError:

    def pairwise(items):  # type: ignore
        return zip(items, items[1:])


class html_objects_list:
    def __init__(self, format_string: str, max_items: int = 10):
        self.format_string = format_string
        self.max_items = max_items

    def __call__(self, fn: Callable) -> Callable:
        @wraps(fn)
        def wrapped(*args, **kwargs):
            results = list(islice(fn(*args, **kwargs), self.max_items))
            return format_html("<br>".join(self.format_string for _ in results), *chain.from_iterable(results))

        return wrapped


@admin.register(UserSession)
class UserSessionAdmin(admin.ModelAdmin):
    list_display = (
        "session_key",
        "user",
        "browser_fingerprints",
        "request_fingerprints",
        "created",
    )
    list_filter = ("created",)
    search_fields = (
        "user__username",
        "session_key",
    )
    ordering = ("-created",)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("user")
            .prefetch_related(
                Prefetch(
                    "browserfingerprints",
                    queryset=BrowserFingerprint.objects.order_by("visitor_id", "-created").distinct("visitor_id"),
                ),
                Prefetch(
                    "requestfingerprints",
                    queryset=RequestFingerprint.objects.order_by("user_session", "user_agent", "-created").distinct(
                        "user_session", "user_agent"
                    ),
                ),
            )
        )

    @html_objects_list('{} &nbsp;&nbsp; <a href="{}"><code>{}</code></a>')
    def browser_fingerprints(self, instance) -> Iterator[tuple]:
        for fingerprint in instance.browserfingerprints.all():
            yield (
                fingerprint.created.date(),
                reverse("admin:fingerprint_browserfingerprint_changelist") + f"?user_session={instance.id}",
                fingerprint.get_value_display(),
            )

    @html_objects_list('{} &nbsp;&nbsp; <a href="{}"><code>{}</code></a>')
    def request_fingerprints(self, instance) -> Iterator[tuple]:
        for fingerprint in instance.requestfingerprints.all():
            yield (
                fingerprint.created.date(),
                reverse("admin:fingerprint_requestfingerprint_changelist") + f"?user_session={instance.id}",
                fingerprint.get_value_display(),
            )


class FingerprintBaseAdmin(admin.ModelAdmin):
    list_display: tuple[str, ...] = (
        "id",
        "user_session",
        "get_user",
    )
    list_filter: tuple[str, ...] = ("created",)
    ordering: tuple[str, ...] = ("-created",)
    search_fields: tuple[str, ...] = (
        "user_session__session_key",
        "user_session__user__email",
        "user_session__user__username",
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("user_session__user")

    @admin.display(description="user")
    def get_user(self, instance):
        return instance.user_session.user


@admin.register(BrowserFingerprint)
class BrowserFingerprintAdmin(FingerprintBaseAdmin):
    list_display = (
        *FingerprintBaseAdmin.list_display,
        "visitor_id",
        "created",
    )
    search_fields = (
        "visitor_id",
        *FingerprintBaseAdmin.search_fields,
    )


@admin.register(RequestFingerprint)
class RequestFingerprintAdmin(FingerprintBaseAdmin):
    list_display = (
        *FingerprintBaseAdmin.list_display,
        "ip",
        "user_agent",
        "created",
    )

    search_fields = (
        "ip",
        "user_agent",
        "accept",
        "content_encoding",
        "content_language",
        *FingerprintBaseAdmin.search_fields,
    )


class NumFingerprintsListFilter(admin.SimpleListFilter):
    groups = list(pairwise((1, 4, 6, 11, 21, 10_001)))
    query_parameter: str

    def lookups(self, request, model_admin):
        return [(i, f"{min_}-{max_ - 1}") for i, (min_, max_) in enumerate(self.groups)]

    def queryset(self, request, queryset):
        if not (value := self.value()):
            return

        try:
            min_, max_ = self.groups[int(value)]
        except (KeyError, ValueError):
            return

        return queryset.filter(
            **{
                f"{self.query_parameter}__gte": min_,
                f"{self.query_parameter}__lt": max_,
            }
        )


class NumBrowserFingerprintsListFilter(NumFingerprintsListFilter):
    title = "unique browser fingerprints"
    parameter_name = "num_browser_fingerprints_group"
    query_parameter = "num_browser_fingerprints"


class NumRequestFingerprintsListFilter(NumFingerprintsListFilter):
    title = "unique request fingerprints"
    parameter_name = "num_request_fingerprints_group"
    query_parameter = "num_request_fingerprints"


class LastFingerprintCreatedListFilter(admin.SimpleListFilter):
    title = "last fingerprint capture"
    parameter_name = "fingerprint_captured_days"

    def lookups(self, request, model_admin):
        return [
            (1, "1 day"),
            (7, "1 week"),
            (30, "1 month"),
            (90, "3 months"),
            (365, "12 months"),
        ]

    def queryset(self, request, queryset):
        if not (value := self.value()):
            return

        try:
            days = timedelta(days=int(value))
        except ValueError:
            return

        since = now() - days
        return queryset.annotate(
            last_browser_fingerprint=Max("sessions__browserfingerprints__created"),
            last_request_fingerprint=Max("sessions__requestfingerprints__created"),
        ).filter(Q(last_browser_fingerprint__gte=since) | Q(last_request_fingerprint__gte=since))


@admin.register(UserFingerprint)
class UserFingerprintAdmin(admin.ModelAdmin):
    list_display = (
        "username",
        "email",
        "sessions",
        "num_browser_fingerprints",
        "browser_fingerprints",
        "num_request_fingerprints",
        "request_fingerprints",
    )
    list_filter = (
        LastFingerprintCreatedListFilter,
        NumBrowserFingerprintsListFilter,
        NumRequestFingerprintsListFilter,
    )
    search_fields = (
        "username",
        "email",
    )

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related(
                Prefetch("sessions", queryset=UserSession.objects.all().order_by("-created")),
                Prefetch(
                    "sessions__browserfingerprints",
                    queryset=BrowserFingerprint.objects.order_by("visitor_id", "-created").distinct("visitor_id"),
                ),
                Prefetch(
                    "sessions__requestfingerprints",
                    queryset=RequestFingerprint.objects.order_by("user_session", "user_agent", "-created").distinct(
                        "user_session", "user_agent"
                    ),
                ),
            )
            .annotate(
                num_browser_fingerprints=Count("sessions__browserfingerprints__visitor_id", distinct=True),
                num_request_fingerprints=Count("sessions__requestfingerprints__user_agent", distinct=True),
            )
        )

    @admin.display(description="#")
    def num_browser_fingerprints(self, instance):
        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:fingerprint_browserfingerprint_changelist") + f"?q={instance.email}",
            instance.num_browser_fingerprints,
        )

    @admin.display(description="#")
    def num_request_fingerprints(self, instance):
        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:fingerprint_requestfingerprint_changelist") + f"?q={instance.email}",
            instance.num_request_fingerprints,
        )

    @admin.display(description="last sessions")
    @html_objects_list('{} &nbsp;&nbsp; <a href="{}"><code>{}</code></a>')
    def sessions(self, instance) -> Iterator[tuple]:
        for session in instance.sessions.all():
            yield (
                session.created.date(),
                reverse("admin:fingerprint_usersession_changelist") + f"?id={instance.id}",
                session.get_value_display(),
            )

    @admin.display(description="last browsers")
    @html_objects_list('{} &nbsp;&nbsp; <a href="{}"><code>{}</code></a>')
    def browser_fingerprints(self, instance) -> Iterator[tuple]:
        for session in instance.sessions.all():
            for fingerprint in session.browserfingerprints.all():
                yield (
                    fingerprint.created.date(),
                    reverse("admin:fingerprint_browserfingerprint_changelist") + f"?id={fingerprint.id}",
                    fingerprint.get_value_display(),
                )

    @admin.display(description="last requests")
    @html_objects_list('{} &nbsp;&nbsp; <a href="{}"><code>{}</code></a>')
    def request_fingerprints(self, instance) -> Iterator[tuple]:
        for session in instance.sessions.all():
            for fingerprint in session.requestfingerprints.all():
                yield (
                    fingerprint.created.date(),
                    reverse("admin:fingerprint_requestfingerprint_changelist") + f"?id={fingerprint.id}",
                    fingerprint.get_value_display(),
                )



"""
Django settings used in tests.
"""
# cookiecutter-rt-pkg macro: requires cookiecutter.is_django_package

DEBUG = True
SECRET_KEY = "DUMMY"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = __name__
urlpatterns = []  # type: ignore



from django.db import models


class TruncatedCharField(models.CharField):
    def pre_save(self, model_instance, add):
        value = getattr(model_instance, self.attname)
        if value and len(value) > self.max_length:
            # Silently truncate to fit the DB constraint
            value = value[: self.max_length]
            setattr(model_instance, self.attname, value)
        return value

    def get_db_prep_value(self, value, connection, prepared=False):
        if value and len(value) > self.max_length:
            # Silently truncate to fit the DB constraint
            value = value[: self.max_length]
        return super().get_db_prep_value(value, connection, prepared)



# Generated by Django 4.2.4 on 2023-08-10 14:51

import django.db.models.deletion
from django.db import migrations, models


def add_default_url(apps, schema_editor):
    db_alias = schema_editor.connection.alias

    Url = apps.get_model("fingerprint", "Url")
    default_url = Url.objects.using(db_alias).create(value="?")

    RequestFingerprint = apps.get_model("fingerprint", "RequestFingerprint")
    RequestFingerprint.objects.using(db_alias).update(url=default_url)

    BrowserFingerprint = apps.get_model("fingerprint", "BrowserFingerprint")
    BrowserFingerprint.objects.using(db_alias).update(url=default_url)


class Migration(migrations.Migration):
    dependencies = [
        (
            "fingerprint",
            "0002_url_browserfingerprint_fingerprint_url_id_9faee4_idx_and_more",
        ),
    ]

    operations = [
        migrations.RunPython(add_default_url, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="browserfingerprint",
            name="url",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(model_name)ss",
                to="fingerprint.url",
            ),
        ),
        migrations.AlterField(
            model_name="requestfingerprint",
            name="url",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(model_name)ss",
                to="fingerprint.url",
            ),
        ),
    ]



"""
Test setup tests.

This test file is here always to indicate that everything was installed and the CI was able to run tests.
It should always pass as long as all dependencies are properly installed.
"""

from datetime import datetime, timedelta

import pytest
from freezegun import freeze_time


def test__setup():
    with freeze_time(datetime.now() - timedelta(days=1)):
        pass

    with pytest.raises(ZeroDivisionError):
        1 / 0



# Generated by Django 4.2.20 on 2025-03-12 11:44

from django.db import migrations, models


def fill_referer(apps, schema_editor):
    UserSession = apps.get_model("fingerprint", "UserSession")
    RequestFingerprint = apps.get_model("fingerprint", "RequestFingerprint")

    # fill referer value from the earliest session request fingerprint
    fingerprint_subquery = (
        RequestFingerprint.objects.filter(user_session_id=models.OuterRef("pk"))
        .order_by("created")
        .values_list("referer", flat=True)[:1]
    )

    UserSession.objects.annotate(fp_referer=models.Subquery(fingerprint_subquery)).filter(
        fp_referer__isnull=False
    ).update(referer=models.F("fp_referer"))


class Migration(migrations.Migration):
    dependencies = [
        ("fingerprint", "0005_user_session_utm_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="usersession",
            name="referer",
            field=models.CharField(blank=True, max_length=2047),
        ),
        migrations.RunPython(fill_referer, migrations.RunPython.noop),
    ]



# Generated by Django 4.2 on 2023-05-02 10:22

import django.contrib.auth.models
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="UserFingerprint",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("auth.user",),
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name="UserSession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("session_key", models.CharField(max_length=40)),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sessions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RequestFingerprint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("ip", models.GenericIPAddressField(blank=True, null=True)),
                ("user_agent", models.CharField(blank=True, max_length=255)),
                ("accept", models.CharField(blank=True, max_length=255)),
                ("content_encoding", models.CharField(blank=True, max_length=255)),
                ("content_language", models.CharField(blank=True, max_length=255)),
                (
                    "user_session",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(model_name)ss",
                        to="fingerprint.usersession",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="BrowserFingerprint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("visitor_id", models.CharField(max_length=255)),
                (
                    "user_session",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(model_name)ss",
                        to="fingerprint.usersession",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddIndex(
            model_name="usersession",
            index=models.Index(fields=["user", "-created"], name="fingerprint_user_id_b99b01_idx"),
        ),
        migrations.AddConstraint(
            model_name="usersession",
            constraint=models.UniqueConstraint(fields=("session_key", "user"), name="unique_user_session"),
        ),
        migrations.AddIndex(
            model_name="requestfingerprint",
            index=models.Index(
                fields=["user_session", "-created"],
                name="fingerprint_user_se_1d7864_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="requestfingerprint",
            index=models.Index(fields=["ip", "-created"], name="fingerprint_ip_56a800_idx"),
        ),
        migrations.AddIndex(
            model_name="requestfingerprint",
            index=models.Index(fields=["user_agent", "-created"], name="fingerprint_user_ag_5d70f2_idx"),
        ),
        migrations.AddIndex(
            model_name="browserfingerprint",
            index=models.Index(
                fields=["user_session", "-created"],
                name="fingerprint_user_se_480add_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="browserfingerprint",
            index=models.Index(fields=["visitor_id", "-created"], name="fingerprint_visitor_af9597_idx"),
        ),
    ]



from __future__ import annotations

import typing
from collections import Counter
from contextlib import suppress

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sessions.models import Session
from django.db import IntegrityError, models, transaction
from django.db.models import Count, Model
from django.db.models.signals import post_save
from django.dispatch import receiver

from .fields import TruncatedCharField

if typing.TYPE_CHECKING:
    from django.contrib.auth.models import AbstractBaseUser
    from django.http import HttpRequest
    from django.shortcuts import SupportsGetAbsoluteUrl


class UserSession(models.Model):
    # by default, django stores session data in database; however, we cannot rely on it,
    # since other session backends may be used, plus django doesn't store user-session
    # mapping, so we need to have additional table for that

    user: models.ForeignKey = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, blank=True, null=True, related_name="sessions"
    )
    session_key: models.CharField = models.CharField(max_length=40)
    created: models.DateTimeField = models.DateTimeField(auto_now_add=True)

    referer: models.CharField = TruncatedCharField(max_length=2047, blank=True)

    # Urchin Tracking Module (UTM) parameters
    utm_source: models.CharField = TruncatedCharField(max_length=255, blank=True)
    utm_medium: models.CharField = TruncatedCharField(max_length=255, blank=True)
    utm_campaign: models.CharField = TruncatedCharField(max_length=255, blank=True)
    utm_content: models.CharField = TruncatedCharField(max_length=255, blank=True)
    utm_term: models.CharField = TruncatedCharField(max_length=255, blank=True)

    def __str__(self):
        return self.session_key

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["session_key", "user"], name="unique_user_session"),
        ]
        indexes = [
            models.Index(fields=["user", "-created"]),
        ]

    def get_value_display(self) -> str:
        return self.session_key[:8]


@receiver(post_save, sender=Session)
def connect_user_to_session(sender, instance, created, **kwargs):
    if user_id := instance.get_decoded().get("_auth_user_id"):
        UserSession.objects.update_or_create(
            session_key=instance.session_key,
            defaults=dict(user_id=user_id),
        )


class UrlQuerySet(models.QuerySet):
    def get_get_or_create(self, defaults: dict = {}, **query) -> tuple[Model, bool]:
        """
        Try to retrieve an object using simple `get()` query, and fallback to `get_or_create()`
        if the object is not found.

        This is required for django-cacheops to be able to cache the `get()` query, since it
        cannot cache `get_or_create` queries.
        """
        with suppress(self.model.DoesNotExist):
            return self.get(**query), False

        with transaction.atomic():
            try:
                return self.create(**{**query, **defaults}), True
            except IntegrityError:
                return self.get(**query), False


class Url(models.Model):
    value = TruncatedCharField(max_length=2048, unique=True)

    objects = UrlQuerySet.as_manager()

    def __str__(self) -> str:
        return self.value


class AbstractFingerprint(models.Model):
    user_session: models.ForeignKey = models.ForeignKey(
        UserSession, on_delete=models.CASCADE, related_name="%(model_name)ss"
    )
    url: models.ForeignKey = models.ForeignKey(Url, on_delete=models.CASCADE, related_name="%(model_name)ss")
    created: models.DateTimeField = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=["user_session", "-created"]),
            models.Index(fields=["url", "user_session"]),
        ]

    @property
    def user(self) -> AbstractBaseUser | None:
        return self.user_session.user

    @classmethod
    def get_count_for_urls(cls, urls: list[str]) -> Counter[str]:
        existing_urls = dict(Url.objects.filter(value__in=urls).values_list("id", "value"))

        # this is SELECT COUNT(*) GROUP BY in django:
        ids_and_hits = (
            cls.objects.filter(url__in=existing_urls)
            .values("url")
            .annotate(hits=Count("user_session", distinct=True))
            .order_by("url")
            .values_list("url", "hits")
        )

        counter = Counter({existing_urls[id_]: hits for id_, hits in ids_and_hits})

        max_length = Url._meta.get_field("value").max_length
        counter.update({url: counter[url[:max_length]] for url in urls if len(url) > max_length})
        return counter

    @classmethod
    def get_count_for_objects(
        cls, request: HttpRequest, objects: list[SupportsGetAbsoluteUrl]
    ) -> Counter[SupportsGetAbsoluteUrl]:
        url_to_object = {request.build_absolute_uri(obj.get_absolute_url()): obj for obj in objects}
        counter = cls.get_count_for_urls(set(url_to_object.keys()))
        return Counter({obj: counter[url] for url, obj in url_to_object.items()})


class BrowserFingerprint(AbstractFingerprint):
    visitor_id: models.CharField = models.CharField(max_length=255)

    def __str__(self) -> str:
        return self.visitor_id

    class Meta(AbstractFingerprint.Meta):  # noqa: D106
        indexes = [
            *AbstractFingerprint.Meta.indexes,
            models.Index(fields=["visitor_id", "-created"]),
        ]

    def get_value_display(self) -> str:
        return self.visitor_id[:8]


class RequestFingerprint(AbstractFingerprint):
    ip: models.GenericIPAddressField = models.GenericIPAddressField(blank=True, null=True)
    user_agent: models.CharField = TruncatedCharField(max_length=255, blank=True)
    accept: models.CharField = TruncatedCharField(max_length=255, blank=True)
    content_encoding: models.CharField = TruncatedCharField(max_length=255, blank=True)
    content_language: models.CharField = TruncatedCharField(max_length=255, blank=True)
    referer: models.CharField = TruncatedCharField(max_length=2047, blank=True)
    cf_ipcountry: models.CharField = TruncatedCharField(max_length=16, blank=True)

    def __str__(self):
        return f"{self.ip} {self.user_agent}"

    class Meta(AbstractFingerprint.Meta):  # noqa: D106
        indexes = [
            *AbstractFingerprint.Meta.indexes,
            models.Index(fields=["ip", "-created"]),
            models.Index(fields=["user_agent", "-created"]),
        ]

    def get_value_display(self) -> str:
        return self.user_agent[:24] + "..."


class UserFingerprint(get_user_model()):  # type: ignore
    """Proxy model for admin site, since django doesn't allow to register two admins for the same model."""

    class Meta:  # noqa: D106
        proxy = True



# Generated by Django 4.2.20 on 2025-03-11 19:52

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("fingerprint", "0004_requestfingerprint_referer_country"),
    ]

    operations = [
        migrations.AddField(
            model_name="usersession",
            name="utm_campaign",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name="usersession",
            name="utm_content",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name="usersession",
            name="utm_medium",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name="usersession",
            name="utm_source",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name="usersession",
            name="utm_term",
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-10 14:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("fingerprint", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Url",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.CharField(max_length=255, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name="browserfingerprint",
            name="url",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(model_name)ss",
                to="fingerprint.url",
            ),
        ),
        migrations.AddField(
            model_name="requestfingerprint",
            name="url",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(model_name)ss",
                to="fingerprint.url",
            ),
        ),
        migrations.AddIndex(
            model_name="browserfingerprint",
            index=models.Index(fields=["url", "user_session"], name="fingerprint_url_id_9faee4_idx"),
        ),
        migrations.AddIndex(
            model_name="requestfingerprint",
            index=models.Index(fields=["url", "user_session"], name="fingerprint_url_id_6c4d95_idx"),
        ),
    ]



# Generated by Django 6.0.2 on 2026-02-05 10:39

from django.db import migrations

import fingerprint.fields


class Migration(migrations.Migration):
    dependencies = [
        ("fingerprint", "0006_usersession_referer"),
    ]

    operations = [
        migrations.AlterField(
            model_name="url",
            name="value",
            field=fingerprint.fields.TruncatedCharField(max_length=2048, unique=True),
        ),
    ]
